/*
 *  xfvnc
 *
 *  Copyright (c) 2007 Brian Tarricone <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License ONLY.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <gtk/gtk.h>
#include <glade/glade.h>

#include <libxfce4util/libxfce4util.h>
#include <libxfcegui4/libxfcegui4.h>

#include <vncdisplay.h>

typedef struct
{
    gchar *hostname;
    guint display_num;
    guint32 view_only:1,
            allow_other:1,
            fullscreen:1;
} XfvncRecentServer;

static void
vnc_connected(VncDisplay *vnc,
              gpointer user_data)
{
    GladeXML *gxml = user_data;
    GtkWidget *lbl_progress;
    
    lbl_progress = glade_xml_get_widget(gxml, "progress_label");
    gtk_label_set_text(GTK_LABEL(lbl_progress),
                       _("Connected.  Starting session..."));
}

static void
vnc_initialized(VncDisplay *vnc,
                gpointer user_data)
{
    GladeXML *gxml = user_data;
    GtkWidget *win, *connect_win, *progress_dlg, *chk_fullscreen;
    guint pulse_id;
    gboolean fullscreen;
    gchar *title;
    
    progress_dlg = glade_xml_get_widget(gxml, "progress_dlg");
    pulse_id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(progress_dlg),
                                "pulse-timeout-id"));
    g_source_remove(pulse_id);
    gtk_widget_destroy(progress_dlg);
    
    chk_fullscreen = glade_xml_get_widget(gxml, "chk_fullscreen");
    fullscreen = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(chk_fullscreen));
    
    connect_win = glade_xml_get_widget(gxml, "connect_window");
    gtk_widget_destroy(connect_win);
    
    win = glade_xml_get_widget(gxml, "vnc_window");
    title = g_strdup_printf(_("XfVNC .:. %s"), vnc_display_get_name(vnc));
    gtk_window_set_title(GTK_WINDOW(win), title);
    g_free(title);
    gtk_widget_show_all(win);
    
    if(fullscreen)
        gtk_window_fullscreen(GTK_WINDOW(win));
}

static void
vnc_auth_failure(VncDisplay *vnc,
                 const gchar *message,
                 gpointer user_data)
{
    GladeXML *gxml = user_data;
    GtkWidget *progress_dlg = glade_xml_get_widget(gxml, "progress_dlg");
    
    xfce_message_dialog(GTK_WINDOW(progress_dlg), _("Connection Error"),
                        GTK_STOCK_DIALOG_ERROR,
                        _("There was an error connecting to the VNC server"),
                        message ? message : _("Unfortunately, the type of error is unknown."),
                        GTK_STOCK_QUIT, GTK_RESPONSE_ACCEPT, NULL);
    
    gtk_main_quit();
}

static void
vnc_desktop_resize(VncDisplay *vnc,
                   gint width,
                   gint height,
                   gpointer user_data)
{
    GladeXML *gxml = user_data;
    GtkWidget *win;
    
    win = glade_xml_get_widget(gxml, "vnc_window");
    gtk_window_resize(GTK_WINDOW(win), width, height);
}

static gboolean
progress_pulse(gpointer data)
{
    GtkProgressBar *pbar = data;
    gtk_progress_bar_pulse(pbar);
    return TRUE;
}

static void
progress_cancel_clicked(GtkWidget *widget,
                        gpointer user_data)
{
    GladeXML *gxml = user_data;
    GtkWidget *progress_dlg, *win, *vnc, *btn;
    guint pulse_id;
    
    progress_dlg = glade_xml_get_widget(gxml, "progress_dlg");
    pulse_id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(progress_dlg),
                                                  "pulse-timeout-id"));
    g_source_remove(pulse_id);
    gtk_widget_hide(progress_dlg);
    
    win = glade_xml_get_widget(gxml, "vnc_window");
    vnc = g_object_get_data(G_OBJECT(win), "vnc-display");
    g_object_unref(G_OBJECT(vnc));
    gtk_widget_destroy(vnc);
    g_object_set_data(G_OBJECT(win), "vnc-display", NULL);
    gtk_widget_hide(win);
    
    btn = glade_xml_get_widget(gxml, "btn_progress_cancel");
    g_signal_handlers_disconnect_by_func(G_OBJECT(btn),
                                         G_CALLBACK(progress_cancel_clicked),
                                         gxml);
}

static void
connect_clicked(GtkWidget *widget,
                gpointer user_data)
{
    GladeXML *gxml = user_data;
    GtkWidget *win, *sw, *vnc, *entry_host, *entry_displaynum, *entry_password;
    GtkWidget *progress_dlg;
    gchar *host, *dnum, *pass, *port;
    guint pulse_id;
    
    entry_host = glade_xml_get_widget(gxml, "entry_host");
    host = gtk_editable_get_chars(GTK_EDITABLE(entry_host), 0, -1);
    entry_displaynum = glade_xml_get_widget(gxml, "entry_displaynum");
    dnum = gtk_editable_get_chars(GTK_EDITABLE(entry_displaynum), 0, -1);
    entry_password = glade_xml_get_widget(gxml, "entry_password");
    pass = gtk_editable_get_chars(GTK_EDITABLE(entry_password), 0, -1);
    
    port = g_strdup_printf("%d", 5900 + atoi(dnum));
    g_free(dnum);
    
    win = glade_xml_get_widget(gxml, "vnc_window");
    gtk_widget_realize(win);
    
    sw = glade_xml_get_widget(gxml, "vnc_window_sw");
    gtk_widget_show(sw);
    
    vnc = vnc_display_new();
    vnc_display_set_use_shm(VNC_DISPLAY(vnc), TRUE);
    vnc_display_set_pointer_local(VNC_DISPLAY(vnc), TRUE);
    vnc_display_set_pointer_grab(VNC_DISPLAY(vnc), FALSE);
    vnc_display_set_keyboard_grab(VNC_DISPLAY(vnc), FALSE);
    vnc_display_set_credential(VNC_DISPLAY(vnc),
                               VNC_DISPLAY_CREDENTIAL_CLIENTNAME,
                               "XfVNC");
    vnc_display_set_credential(VNC_DISPLAY(vnc),
                               VNC_DISPLAY_CREDENTIAL_PASSWORD,
                               pass);
    vnc_display_open_host(VNC_DISPLAY(vnc), host, port);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), vnc);
    gtk_widget_realize(vnc);
    g_object_set_data(G_OBJECT(win), "vnc-display",
                      g_object_ref(G_OBJECT(vnc)));
    g_signal_connect(G_OBJECT(vnc), "vnc-connected",
                     G_CALLBACK(vnc_connected), gxml);
    g_signal_connect(G_OBJECT(vnc), "vnc-initialized",
                     G_CALLBACK(vnc_initialized), gxml);
    g_signal_connect(G_OBJECT(vnc), "vnc-disconnected",
                     G_CALLBACK(gtk_main_quit), NULL);
    g_signal_connect(G_OBJECT(vnc), "vnc-auth-failure",
                     G_CALLBACK(vnc_auth_failure),gxml);
    g_signal_connect(G_OBJECT(vnc), "vnc-desktop-resize",
                     G_CALLBACK(vnc_desktop_resize), gxml);
    g_signal_connect(G_OBJECT(vnc), "vnc-keyboard-grab",
                     G_CALLBACK(vnc_display_set_keyboard_grab),
                     GUINT_TO_POINTER(FALSE));
    g_signal_connect(G_OBJECT(vnc), "vnc-pointer-grab",
                     G_CALLBACK(vnc_display_set_pointer_grab),
                     GUINT_TO_POINTER(FALSE));
    
    g_free(host);
    g_free(port);
    g_free(pass);
    
    progress_dlg = glade_xml_get_widget(gxml, "progress_dlg");
    gtk_widget_show_all(progress_dlg);
    g_signal_connect(G_OBJECT(glade_xml_get_widget(gxml,
                                                   "btn_progress_cancel")),
                     "clicked", G_CALLBACK(progress_cancel_clicked), gxml);
    
    pulse_id = g_timeout_add(150, progress_pulse,
                             glade_xml_get_widget(gxml, "pbar"));
    g_object_set_data(G_OBJECT(progress_dlg), "pulse-timeout-id",
                      GUINT_TO_POINTER(pulse_id));
}

static void
make_text_combo(GtkComboBox *combo)
{
    GtkCellRenderer *render;
    GtkListStore *ls;
    
    ls = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_POINTER);
    gtk_combo_box_set_model(combo, GTK_TREE_MODEL(ls));
    g_object_unref(G_OBJECT(ls));
    
    render = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), render, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), render,
                                   "text", 0,
                                   NULL);
}

/* yes, i know i could use GtkRecentManager and whatnot, but this is
 * just as easy and doesn't require a really recent gtk */
static gboolean
load_recent_servers(GtkComboBox *combo)
{
    gboolean ret = FALSE;
    gchar *recent_file = NULL, groupbuf[128], *fullhost;
    XfceRc *rcfile = NULL;
    XfvncRecentServer *rs;
    gint nrecent, i;
    GtkListStore *ls;
    
    recent_file = xfce_resource_save_location(XFCE_RESOURCE_CONFIG,
                                              "xfvnc/xfvnc-history.rc",
                                              FALSE);
    if(!recent_file)
        goto out;
    
    rcfile = xfce_rc_simple_open(recent_file, TRUE);
    if(!rcfile)
        goto out;
    
    if(!xfce_rc_has_group(rcfile, "global"))
        goto out;
    xfce_rc_set_group(rcfile, "global");
    nrecent = xfce_rc_read_int_entry(rcfile, "nrecent", 0);
    
    ls = GTK_LIST_STORE(gtk_combo_box_get_model(combo));
    
    for(i = 0; i < nrecent; ++i) {
        GtkTreeIter iter;
        
        g_snprintf(groupbuf, sizeof(groupbuf), "recent%d", i);
        
        if(!xfce_rc_has_group(rcfile, groupbuf))
            continue;
        xfce_rc_set_group(rcfile, groupbuf);
        
        rs = g_new0(XfvncRecentServer, 1);
        rs->hostname = g_strdup(xfce_rc_read_entry(rcfile, "hostname", NULL));
        if(!rs->hostname || !*rs->hostname) {
            g_free(rs->hostname);
            g_free(rs);
            continue;
        }
        rs->display_num = xfce_rc_read_int_entry(rcfile, "display_num", 0);
        rs->view_only = xfce_rc_read_bool_entry(rcfile, "view_only", FALSE);
        rs->allow_other = xfce_rc_read_bool_entry(rcfile, "allow_other", FALSE);
        rs->fullscreen = xfce_rc_read_bool_entry(rcfile, "fullscreen", FALSE);
        
        fullhost = g_strdup_printf("%s:%d", rs->hostname, rs->display_num);
        
        gtk_list_store_append(ls, &iter);
        gtk_list_store_set(ls, &iter,
                           0, fullhost,
                           1, rs,
                           -1);
        g_free(fullhost);
    }
    
out:
    if(rcfile)
        xfce_rc_close(rcfile);
    g_free(recent_file);
    
    return ret;
}

static void
recent_combo_changed(GtkComboBox *combo,
                     gpointer user_data)
{
    GladeXML *gxml = user_data;
    GtkTreeIter iter;
    GtkTreeModel *model;
    XfvncRecentServer *rs = NULL;
    gchar dispbuf[32];
    
    if(!gtk_combo_box_get_active_iter(combo, &iter))
        return;
    
    model = gtk_combo_box_get_model(combo);
    gtk_tree_model_get(model, &iter,
                       1, &rs,
                       -1);
    if(!rs)
        return;
    
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(gxml, "entry_host")),
                       rs->hostname);
    
    g_snprintf(dispbuf, sizeof(dispbuf), "%d", rs->display_num);
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(gxml,
                                                      "entry_displaynum")),
                       dispbuf);
    
    gtk_entry_set_text(GTK_ENTRY(glade_xml_get_widget(gxml, "entry_password")),
                       "");
    
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(gxml,
                                                                        "chk_viewonly")),
                                 rs->view_only);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(gxml,
                                                                        "chk_allowother")),
                                 rs->allow_other);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget(gxml,
                                                                        "chk_fullscreen")),
                                 rs->fullscreen);
}

int
main(int argc,
     char **argv)
{
    GladeXML *gxml;
    gchar *glade_file;
    GtkWidget *connect_window, *combo_recent, *entry_host;
    
    xfce_textdomain(GETTEXT_PACKAGE, LOCALEDIR, "UTF-8");
    
    gtk_init(&argc, &argv);
    
    glade_file = xfce_resource_lookup(XFCE_RESOURCE_DATA,
                                      "xfvnc/xfvnc.glade");
    if(!glade_file)
        glade_file = g_strdup("xfvnc.glade");
    if(!glade_file || !(gxml = glade_xml_new(glade_file, NULL, NULL))) {
        g_error("Unable to load xfvnc.glade");
        return 1;
    }
    g_free(glade_file);
    
    glade_xml_signal_autoconnect(gxml);
    
    connect_window = glade_xml_get_widget(gxml, "connect_window");
    gtk_widget_show_all(connect_window);
    g_signal_connect(G_OBJECT(glade_xml_get_widget(gxml, "btn_connect")),
                     "clicked", G_CALLBACK(connect_clicked), gxml);
    
    combo_recent = glade_xml_get_widget(gxml, "combo_recent");
    make_text_combo(GTK_COMBO_BOX(combo_recent));
    load_recent_servers(GTK_COMBO_BOX(combo_recent));
    g_signal_connect(G_OBJECT(combo_recent), "changed",
                     G_CALLBACK(recent_combo_changed), gxml);
    
    entry_host = glade_xml_get_widget(gxml, "entry_host");
    gtk_widget_grab_focus(entry_host);
    
    gtk_main();
    
    g_object_unref(G_OBJECT(gxml));
    gxml = NULL;
    
    return 0;
}
